/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace scripts
{

//================== console.lua ==================
static const unsigned char temp_binary_data_4[] =
"--- Console init.\n"
"-- Runs when the console is loaded in the GUI. This anonymous script sets global\n"
"-- variables, so be careful if you use it directly.\n"
"-- @script console\n"
"-- @pragma nostrip\n"
"-- @type Anonymous\n"
"\n"
"io      = require ('io')\n"
"object  = require ('el.object')\n"
"command = require ('el.command')\n"
"script  = require ('el.script')\n"
"\n"
"console = {\n"
"    --- Log to stdout.\n"
"    -- Calls `tostring` on each argument then prints the combined result to \n"
"    -- stdout\n"
"    -- @function console.log\n"
"    -- @param ... Things to log\n"
"    log = function (...)\n"
"        local out = \"\"\n"
"        for i = 1, select ('#', ...) do\n"
"            out = out .. tostring (select (i, ...)) .. \"\\t\"\n"
"        end\n"
"        if string.len(out) > 0 then\n"
"            io.stdout:write (out .. \"\\n\")\n"
"        end\n"
"    end\n"
"}\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* console_lua = (const char*) temp_binary_data_4;
}
