// Copyright 2023 Kushview, LLC <info@kushview.net>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

namespace element {

/** Returns true if caps lock is currently down */
bool isCapsLockOn();

} // namespace element
