/* =========================================================================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#pragma once

namespace scripts
{
    extern const char*   amp_lua;
    const int            amp_luaSize = 1467;

    extern const char*   ampui_lua;
    const int            ampui_luaSize = 1690;

    extern const char*   channelize_lua;
    const int            channelize_luaSize = 1861;

    extern const char*   commands_lua;
    const int            commands_luaSize = 1231;

    extern const char*   console_lua;
    const int            console_luaSize = 881;

    extern const char*   content_lua;
    const int            content_luaSize = 336;

    extern const char*   spontonchordchooser_lua;
    const int            spontonchordchooser_luaSize = 5951;

    extern const char*   view_lua;
    const int            view_luaSize = 672;

    // Number of elements in the namedResourceList and originalFileNames arrays.
    const int namedResourceListSize = 8;

    // Points to the start of a list of resource names.
    extern const char* namedResourceList[];

    // Points to the start of a list of resource filenames.
    extern const char* originalFilenames[];

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding data and its size (or a null pointer if the name isn't found).
    const char* getNamedResource (const char* resourceNameUTF8, int& dataSizeInBytes);

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding original, non-mangled filename (or a null pointer if the name isn't found).
    const char* getNamedResourceOriginalFilename (const char* resourceNameUTF8);
}
