/* =========================================================================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#pragma once

namespace element::res
{
    extern const char*   RobotoRegular_ttf;
    const int            RobotoRegular_ttfSize = 168260;

    extern const char*   icontemplate_png;
    const int            icontemplate_pngSize = 7122;

    extern const char*   icon_png;
    const int            icon_pngSize = 417339;

    extern const char*   power_48x48_png;
    const int            power_48x48_pngSize = 1153;

    extern const char*   acknowledgements_txt;
    const int            acknowledgements_txtSize = 3481;

    extern const char*   developers_txt;
    const int            developers_txtSize = 276;

    // Number of elements in the namedResourceList and originalFileNames arrays.
    const int namedResourceListSize = 6;

    // Points to the start of a list of resource names.
    extern const char* namedResourceList[];

    // Points to the start of a list of resource filenames.
    extern const char* originalFilenames[];

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding data and its size (or a null pointer if the name isn't found).
    const char* getNamedResource (const char* resourceNameUTF8, int& dataSizeInBytes);

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding original, non-mangled filename (or a null pointer if the name isn't found).
    const char* getNamedResourceOriginalFilename (const char* resourceNameUTF8);
}
