/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace scripts
{

//================== view.lua ==================
static const unsigned char temp_binary_data_7[] =
"--- A new View script (experimental)\n"
"-- \n"
"-- Script description.\n"
"-- \n"
"-- @type View\n"
"-- @license GPL-3.0-or-later\n"
"-- @author Your Name\n"
"\n"
"local object = require ('el.object')\n"
"local new = object.new\n"
"\n"
"local function instantiate()\n"
"    local w = new ('el.View')\n"
"    w.paint = function (_, g)\n"
"        g:fillAll(0xff00ff00)\n"
"        g:setColor (0xffffffff)\n"
"        g:drawText (\"Hello World\", w:localBounds())\n"
"    end\n"
"    return w\n"
"end\n"
"\n"
"local function handleGraphChange(_, graph)\n"
"end\n"
"\n"
"return {\n"
"    type = \"View\",\n"
"    instantiate = instantiate,\n"
"    graph_changed = handleGraphChange\n"
"}\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* view_lua = (const char*) temp_binary_data_7;
}
