/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace mods
{

//================== session.lua ==================
static const unsigned char temp_binary_data_5[] =
"--- Session.\n"
"-- High level session dealings in Lua.  This module wraps el.Session class\n"
"-- in an simplified API.\n"
"-- @module el.session\n"
"\n"
"local Context = require ('el.Context')\n"
"local Session = require ('el.Session')\n"
"local Node    = require ('el.Node')\n"
"local Graph   = require ('el.Graph')\n"
"\n"
"local M = {}\n"
"local session = Context.instance():session()\n"
"\n"
"function M.toxmlstring() return session:toXmlString() end\n"
"function M.name()        return session.name end\n"
"\n"
"return M\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* session_lua = (const char*) temp_binary_data_5;
}
