/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace mods
{

//================== color.lua ==================
static const unsigned char temp_binary_data_1[] =
"--- Color operations and utilities.\n"
"-- Utility library for manipulating colors.\n"
"-- @module el.color\n"
"local M = {}\n"
"\n"
"return M\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* color_lua = (const char*) temp_binary_data_1;
}
