/* =========================================================================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#pragma once

namespace mods
{
    extern const char*   AudioBuffer_lua;
    const int            AudioBuffer_luaSize = 439;

    extern const char*   color_lua;
    const int            color_luaSize = 225;

    extern const char*   command_lua;
    const int            command_luaSize = 2007;

    extern const char*   object_lua;
    const int            object_luaSize = 6044;

    extern const char*   script_lua;
    const int            script_luaSize = 1072;

    extern const char*   session_lua;
    const int            session_luaSize = 566;

    extern const char*   strings_lua;
    const int            strings_luaSize = 1996;

    // Number of elements in the namedResourceList and originalFileNames arrays.
    const int namedResourceListSize = 7;

    // Points to the start of a list of resource names.
    extern const char* namedResourceList[];

    // Points to the start of a list of resource filenames.
    extern const char* originalFilenames[];

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding data and its size (or a null pointer if the name isn't found).
    const char* getNamedResource (const char* resourceNameUTF8, int& dataSizeInBytes);

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding original, non-mangled filename (or a null pointer if the name isn't found).
    const char* getNamedResourceOriginalFilename (const char* resourceNameUTF8);
}
