/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace scripts
{

//================== content.lua ==================
static const unsigned char temp_binary_data_5[] =
"--- SysEx MIDI Example.\n"
"--\n"
"-- @script      UI.Content\n"
"-- @type        Content\n"
"-- @license     GPL v3\n"
"-- @author      Michael Fisher\n"
"\n"
"local function instantiate (ctx)\n"
"end\n"
"\n"
"return {\n"
"    type = \"Content\",\n"
"    instantiate = instantiate\n"
"}\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* content_lua = (const char*) temp_binary_data_5;
}
