/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace element::res
{

//================== developers.txt ==================
static const unsigned char temp_binary_data_5[] =
"Chris Ryan (digitalelements)\n"
"David Runge (dvzrv)\n"
"Eliot Akira (eliot-akira)\n"
"Max Hunter (nippoo)\n"
"Michael Forrest (michaelforrest)\n"
"Steve Schow (steveschow)\n"
"Trevor Parscal (trevorparscal)\n"
"iamvs2002\n"
"jatinchowdhury18\n"
"mmoole. (mmoole)\n"
"schnitzeltony\n"
"splisp\n"
"Heath Dutton (heathdutton)\n";

const char* developers_txt = (const char*) temp_binary_data_5;
}
