/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace scripts
{

//================== commands.lua ==================
static const unsigned char temp_binary_data_3[] =
"--- List commands registered in the command manager.\n"
"-- @script commands\n"
"-- @type anonymous\n"
"-- @license GPLv3\n"
"-- @author Michael Fisher\n"
"-- @usage\n"
"-- > script.exec ('commands')\n"
"-- SHOW_ABOUT = 256\n"
"-- SHOW_ALL_PLUGIN_WINDOWS = 265\n"
"-- SHOW_SESSION_CONFIG = 260\n"
"-- RECENTS_CLEAR = 4096\n"
"-- GRAPH_OPEN = 1793\n"
"-- GRAPH_SAVE_AS = 1795\n"
"-- UNDO = 4104\n"
"-- ...\n"
"\n"
"local CMD   = require ('el.CommandManager')\n"
"local strings  = require ('el.strings')\n"
"local format = select (1, ...) or 'constants'\n"
"\n"
"if not strings.isslug (format) then\n"
"    error (\"Invalid format: \" .. tostring (format))\n"
"end\n"
"\n"
"local list = {\n"
"    -- List as constants used with the el.command module\n"
"    constants = function (cmds)\n"
"        for k,v in pairs (cmds) do\n"
"            local out = string.upper (strings.tosnake (k))\n"
"            out = out .. \" = \" .. tostring(v)\n"
"            print (out)\n"
"        end\n"
"    end\n"
"}\n"
"\n"
"local cmds = {}\n"
"\n"
"-- add standard commands\n"
"for _, cmd in ipairs (CMD.standard()) do\n"
"    local str = CMD.tostring (cmd)\n"
"    if string.len(str) > 0 then\n"
"        cmds[str] = cmd        \n"
"    end\n"
"end\n"
"\n"
"if type(list[format]) == 'function' then\n"
"    list[format] (cmds)\n"
"end\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* commands_lua = (const char*) temp_binary_data_3;
}
