// Copyright 2023 Kushview, LLC <info@kushview.net>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

#include <string>

namespace element {

struct PresetInfo {
    std::string ID;
    std::string name;
    std::string format;
    std::string file;
};

} // namespace element
