// Copyright 2023 Kushview, LLC <info@kushview.net>
// SPDX-License-Identifier: GPL-3.0-or-later

#define EL_LUA_AUDIO_BUFFER_32 0
#define EL_LUA_AUDIO_BUFFER_COMPILE 1
#include "./AudioBufferImpl.ipp"
