/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace mods
{

//================== script.lua ==================
static const unsigned char temp_binary_data_4[] =
"--- Script module.\n"
"-- @module el.script\n"
"local M = {}\n"
"\n"
"M.DSP        = 'DSP'\n"
"M.DSPUI      = 'DSPUI'\n"
"M.VIEW       = 'View'\n"
"M.GRAPH_VIEW = 'GraphView'\n"
"M.ANONYMOUS  = 'Anonymous'\n"
"\n"
"--- Load a known script.\n"
"-- @function load\n"
"-- @string path The script to run\n"
"-- @tparam table env The environment to use or _ENV\n"
"-- @treturn table The script's descriptor table\n"
"-- @usage script.load ('scriptname')\n"
"function M.load (path, env)\n"
"    local src, e = package.searchpath (path, package.spath)\n"
"    if src then return loadfile (src, 'bt', env or _ENV) end\n"
"    return nil, e\n"
"end\n"
"\n"
"--- Run a known script.\n"
"-- @function exec\n"
"-- @string path The script to run\n"
"-- @tparam table env The environment to use or _ENV\n"
"-- @tparam any ... Arguments passed to script\n"
"-- @treturn any Return value from script or no value\n"
"-- @usage script.exec ('scriptname')\n"
"function M.exec (path, env, ...)\n"
"    local invoke, err = M.load (path, env)\n"
"    if err then return err end\n"
"    return invoke (...)\n"
"end\n"
"\n"
"return M\n"
"\n"
"-- SPDX-FileCopyrightText: Copyright (C) Kushview, LLC.\n"
"-- SPDX-License-Identifier: GPL-3.0-or-later\n";

const char* script_lua = (const char*) temp_binary_data_4;
}
